/**
 * Copyright (C) 2005 Billy Biggs <vektor@dumbterm.net>.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package ca.vektor.magnify;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;

public class Magnify {
	static int outputWidth = 320, outputHeight = 240, scaleFactor = 8;

	static boolean showGrid, paused, slowMotion, redraw;

	static MenuItem fourTimes, eightTimes, sixteenTimes;

	static Point mouseReference;

	static RGB midPoint;

	public static void updateScale() {
		fourTimes.setSelection(scaleFactor == 4);
		eightTimes.setSelection(scaleFactor == 8);
		sixteenTimes.setSelection(scaleFactor == 16);
	}

	public static Image scaleNearest(Display display, Image src, int factor,
			boolean grid) {
		ImageData data = src.getImageData();
		Image dest = new Image(display, data.width * factor, data.height
				* factor);
		GC gc = new GC(dest);
		Color black = display.getSystemColor(SWT.COLOR_BLACK);
		Color red = display.getSystemColor(SWT.COLOR_RED);
		for (int y = 0; y < data.height; y++) {
			for (int x = 0; x < data.width; x++) {
				int dx = x * factor;
				int dy = y * factor;
				RGB rgb = data.palette.getRGB(data.getPixel(x, y));
				Color color = new Color(display, rgb);
				gc.setBackground(color);
				gc.setForeground(color);
				gc.fillRectangle(dx, dy, factor, factor);
				if (y == data.height / 2 && x == data.width / 2) {
					gc.setForeground(red);
					gc.drawRectangle(dx, dy, factor - 1, factor - 1);
					midPoint = rgb;
				} else if (grid) {
					gc.setForeground(black);
					gc.drawLine(dx, dy, dx + factor, dy);
					gc.drawLine(dx + factor - 1, dy, dx + factor - 1, dy
							+ factor);
				}
				color.dispose();
			}
		}
		gc.dispose();
		return dest;
	}

	public static void main(String[] args) {
		final Display display = new Display();
		final Shell shell = new Shell(display, SWT.SHELL_TRIM & ~SWT.RESIZE);
		shell.setLayout(new GridLayout(2, false));
		shell.setText("Magnify");

		Image image = new Image(display, 320, 240);
		final Label current = new Label(shell, SWT.BORDER);
		current.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		current.setImage(image);

		final Label saved = new Label(shell, SWT.BORDER);
		saved.setLayoutData(new GridData(SWT.FILL, SWT.FILL, false, false));
		saved.setImage(image);

		final Label info = new Label(shell, SWT.RIGHT);
		info.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false, 1, 1));
		info.setText(" ");

		final Label help = new Label(shell, SWT.RIGHT);
		help.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false, 1, 1));
		help.setText("Hold SHIFT to move slowly");

		Menu menu = new Menu(shell, SWT.BAR);
		MenuItem file = new MenuItem(menu, SWT.CASCADE);
		file.setText("&File");
		Menu fileMenu = new Menu(shell, SWT.DROP_DOWN);
		file.setMenu(fileMenu);

		MenuItem hold = new MenuItem(fileMenu, SWT.PUSH);
		hold.setText("&Hold Current Image\tSpacebar");
		hold.setAccelerator(' ');
		hold.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				Image hold = new Image(display, current.getImage(),
						SWT.IMAGE_COPY);
				Image old = saved.getImage();
				saved.setImage(hold);
				if (old != null)
					old.dispose();
			}
		});

		final MenuItem pause = new MenuItem(fileMenu, SWT.PUSH);
		pause.setText("&Pause\tP");
		pause.setAccelerator('P');
		pause.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				paused = !paused;
				pause.setText(paused ? "Un&pause\tP" : "&Pause\tP");
				redraw = true;
			}
		});
		new MenuItem(fileMenu, SWT.SEPARATOR);
		MenuItem quit = new MenuItem(fileMenu, SWT.PUSH);
		quit.setText("&Quit\tCtrl+Q");
		quit.setAccelerator(SWT.CONTROL | 'q');
		quit.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				shell.dispose();
			}
		});
		MenuItem scale = new MenuItem(menu, SWT.CASCADE);
		scale.setText("&Scale");
		Menu scaleMenu = new Menu(shell, SWT.DROP_DOWN);
		scale.setMenu(scaleMenu);
		fourTimes = new MenuItem(scaleMenu, SWT.CHECK);
		fourTimes.setText("x 4\t4");
		fourTimes.setAccelerator('4');
		fourTimes.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				scaleFactor = 4;
				updateScale();
				redraw = true;
			}
		});
		eightTimes = new MenuItem(scaleMenu, SWT.CHECK);
		eightTimes.setText("x 8\t8");
		eightTimes.setAccelerator('8');
		eightTimes.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				scaleFactor = 8;
				updateScale();
				redraw = true;
			}
		});
		sixteenTimes = new MenuItem(scaleMenu, SWT.CHECK);
		sixteenTimes.setText("x 16\t1");
		sixteenTimes.setAccelerator('1');
		sixteenTimes.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				scaleFactor = 16;
				updateScale();
				redraw = true;
			}
		});
		updateScale();
		MenuItem options = new MenuItem(menu, SWT.CASCADE);
		options.setText("&Options");
		Menu optionsMenu = new Menu(shell, SWT.DROP_DOWN);
		options.setMenu(optionsMenu);
		final MenuItem grid = new MenuItem(optionsMenu, SWT.CHECK);
		grid.setText("Show Grid\tG");
		grid.setAccelerator('G');
		grid.setSelection(showGrid);
		grid.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				showGrid = !showGrid;
				grid.setSelection(showGrid);
				redraw = true;
			}
		});
		shell.setMenuBar(menu);

		shell.pack();
		shell.open();

		display.timerExec(50, new Runnable() {
			Point lastLocation;

			public void run() {
				if (shell.isDisposed())
					return;
				Point cursorLocation = display.getCursorLocation();
				if (redraw || (!paused && !cursorLocation.equals(lastLocation))) {
					lastLocation = cursorLocation;
					redraw = false;
					if (slowMotion) {
						cursorLocation.x = mouseReference.x
								+ (cursorLocation.x - mouseReference.x) / 4;
						cursorLocation.y = mouseReference.y
								+ (cursorLocation.y - mouseReference.y) / 4;
					}
					Image image = new Image(display, outputWidth / scaleFactor,
							outputHeight / scaleFactor);
					GC gc = new GC(display);
					gc.copyArea(image, cursorLocation.x
							- (outputWidth / scaleFactor) / 2, cursorLocation.y
							- (outputHeight / scaleFactor) / 2);
					gc.dispose();
					Image result = scaleNearest(display, image, scaleFactor,
							showGrid);
					Image old = current.getImage();
					current.setImage(result);
					if (old != null)
						old.dispose();
					image.dispose();
					info.setText("(" + midPoint.red + ", " + midPoint.green
							+ ", " + midPoint.blue + ")");
				}
				display.timerExec(50, this);
			}
		});

		display.addFilter(SWT.KeyDown, new Listener() {
			public void handleEvent(Event event) {
				if ((event.keyCode & SWT.SHIFT) != 0) {
					slowMotion = true;
					mouseReference = display.getCursorLocation();
				}
			}
		});

		display.addFilter(SWT.KeyUp, new Listener() {
			public void handleEvent(Event event) {
				if ((event.keyCode & SWT.SHIFT) != 0) {
					slowMotion = false;
					Point cursorLocation = display.getCursorLocation();
					cursorLocation.x = mouseReference.x
							+ (cursorLocation.x - mouseReference.x) / 4;
					cursorLocation.y = mouseReference.y
							+ (cursorLocation.y - mouseReference.y) / 4;
					display.setCursorLocation(cursorLocation.x,
							cursorLocation.y);
				}
			}
		});

		while (!shell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}
		display.dispose();
	}
}
